/*
 * Copyright 2020, OpenTelemetry Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.opentelemetry;

import static io.restassured.RestAssured.given;

import io.restassured.http.ContentType;
import io.restassured.response.Response;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import org.awaitility.Awaitility;
import org.junit.ClassRule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;
import org.testcontainers.containers.GenericContainer;
import org.testcontainers.containers.Network;
import org.testcontainers.containers.wait.strategy.HttpWaitStrategy;
import org.testcontainers.containers.wait.strategy.Wait;
import org.testcontainers.utility.MountableFile;

/**
 * Integration test to verify that OpenTelemetry artefacts run in JRE 7.
 *
 * <p>An executable JAR with dependencies containing {@link SendTraceToJaeger} is built prior to
 * executing the tests.
 *
 * <p>A Jaeger-all-in-one container is started, then an Alpine JRE 7 container is started with the
 * executable JAR added to it and executed which will send a trace to the Jaeger instance. The test
 * verifies that the trace is received by Jaeger.
 */
@RunWith(JUnit4.class)
public class JavaSevenCompatibilityIntegrationTest {

  private static final String ARCHIVE_NAME = System.getProperty("archive.name");
  private static final String APP_NAME = "SendTraceToJaeger.jar";

  private static final int QUERY_PORT = 16686;
  private static final int COLLECTOR_PORT = 14250;
  private static final String JAEGER_VERSION = "1.17";
  private static final String SERVICE_NAME = "integration test";
  private static final String JAEGER_HOSTNAME = "jaeger";
  private static final String JAEGER_URL = "http://localhost";

  private static final Network network = Network.SHARED;

  @SuppressWarnings("rawtypes")
  @ClassRule
  public static GenericContainer jaegerContainer =
      new GenericContainer<>("jaegertracing/all-in-one:" + JAEGER_VERSION)
          .withNetwork(network)
          .withNetworkAliases(JAEGER_HOSTNAME)
          .withExposedPorts(COLLECTOR_PORT, QUERY_PORT)
          .waitingFor(new HttpWaitStrategy().forPath("/"));

  @SuppressWarnings("rawtypes")
  @ClassRule
  public static GenericContainer jaegerExampleAppContainer =
      new GenericContainer("openjdk:7u111-jre-alpine")
          .withNetwork(network)
          .withCopyFileToContainer(MountableFile.forHostPath(ARCHIVE_NAME), "/app/" + APP_NAME)
          .withCommand(
              "java",
              "-cp",
              "/app/" + APP_NAME,
              "io.opentelemetry.SendTraceToJaeger",
              JAEGER_HOSTNAME,
              Integer.toString(COLLECTOR_PORT))
          .waitingFor(Wait.forLogMessage(".*Bye.*", 1))
          .dependsOn(jaegerContainer);

  @Test
  public void testJaegerExampleAppIntegration() {
    Awaitility.await()
        .atMost(30, TimeUnit.SECONDS)
        .until(JavaSevenCompatibilityIntegrationTest::assertJaegerHaveTrace);
  }

  private static Boolean assertJaegerHaveTrace() {
    try {
      String url =
          String.format(
              "%s/api/traces?service=%s",
              String.format(JAEGER_URL + ":%d", jaegerContainer.getMappedPort(QUERY_PORT)),
              SERVICE_NAME);
      Response response =
          given()
              .headers("Content-Type", ContentType.JSON, "Accept", ContentType.JSON)
              .when()
              .get(url)
              .then()
              .contentType(ContentType.JSON)
              .extract()
              .response();
      Map<String, String> path = response.jsonPath().getMap("data[0]");
      return path.get("traceID") != null;
    } catch (Exception e) {
      return false;
    }
  }
}
